#!/usr/bin/python
from mpi4py import MPI
from numpy  import array
import sys
import pyuppaal
import cPickle as pickle
import datetime
import random

from messages import *
from successor import SuccessorGenerator

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

#Colors for termination detection
( WHITE, BLACK ) = range(2)

class Node:
    def __init__(self, model, query):
        self.waiting = []
        self.passed_waiting = set()
        
        self.succgen = SuccessorGenerator(model, query)

        self.reqlist = []
        self.control_reqlist = []
        self.computation_reqlist = []
        self.reqhandlers = []
        self.control_reqhandlers = []
        self.computation_reqhandlers = []
        self.start_time = datetime.datetime.now()
        self.terminated = False
        self.has_term_detect_token = False
        self.goal_reached = False

    def homenode(self, state):
        """Calculate the node this state belongs to."""
        try:
            return 1 + hash(state) % (size-1)
        except ZeroDivisionError:
            #Short-circuit for single or dual process case
            return size

    def sendstate(self, state):
        #send state to node
        recvnode = self.homenode(state)
        if rank == recvnode:
            if not state in self.passed_waiting: #and not state in self.waiting:
                self.waiting.append(state)
                self.passed_waiting.add(state)

        else:
            self.sendtoremote(state, recvnode)

    def sendtoremote(self, state, recvnode):
        self.term_detect_color = BLACK
        #print "Sending state, %d -> %d" % (rank, recvnode)
        comm.send(state, dest=recvnode, tag=MSG_STATE)

    def init_termination_detection(self):
        #MSG_TERMINATION_DETECTION {{{
        self.termination_detection_req = array(0, dtype=int)
        self.termination_detection_prequest = comm.Recv_init([self.termination_detection_req, MPI.INT], source=size-1, tag=MSG_TERMINATION_DETECTION)
        self.computation_reqlist += [self.termination_detection_prequest]
        #self.computation_reqhandlers += [self.termination_detection_received]
        #}}}
        #MSG_GOAL_REACHED {{{
        self.goal_reached_req = array(0, dtype=int)
        self.goal_reached_prequest = comm.Recv_init([self.goal_reached_req, MPI.INT], source=MPI.ANY_SOURCE, tag=MSG_GOAL_REACHED)
        self.computation_reqlist += [self.goal_reached_prequest]
        #self.computation_reqhandlers += [self.goal_reached_received]
        MPI.Prequest.Startall(self.computation_reqlist)
        #}}}
        self.term_detect_color = BLACK
        self.master_start_termination_detection()

    def master_start_termination_detection(self):
        tosend = array(self.term_detect_color, dtype=int)
        comm.Bsend([tosend, MPI.INT], dest=(rank+1)%size, tag=MSG_TERMINATION_DETECTION)

    def master_termination_detection(self):
        if self.termination_detection_prequest.Test():
            token = self.termination_detection_req
            #print "master received:", token
    
            if token == WHITE and self.term_detect_color == WHITE:
                self.terminated = True
                print "Master reporting: Termination detected!"
            else:
                self.term_detect_color = WHITE
                self.master_start_termination_detection()
            self.termination_detection_prequest.Start()

    def master_goal_detection(self):
        if self.goal_reached_prequest.Test():
            #self.goal_reached_received()
            self.goal_reached = True
            return True
        return False
    #}}}

    def slave(self): #{{{
        #setup prerequests
        #MSG_STATUS {{{
        self.status_reqs = array(0, dtype=int)
        self.status_prequest = comm.Recv_init([self.status_reqs, MPI.INT], source=0, tag=MSG_STATUS)
        self.control_reqlist += [self.status_prequest]
        self.control_reqhandlers += [self.status]
        #}}}
        #MSG_ABORT {{{
        self.abort_reqs = array(0, dtype=int)
        self.abort_prequest = comm.Recv_init([self.abort_reqs, MPI.INT], source=0, tag=MSG_ABORT)
        self.control_reqlist += [self.abort_prequest]
        self.control_reqhandlers += [self.abort]
        #}}}
        #MSG_STATE {{{
        self.state_req = array([0]*300, dtype='uint8') #XXX
        self.state_prequest = comm.Recv_init([self.state_req, MPI.BYTE], source=MPI.ANY_SOURCE, tag=MSG_STATE)
        self.computation_reqlist += [self.state_prequest]
        self.computation_reqhandlers += [self.state_received]
        #}}}
        #MSG_TERMINATION_DETECTION {{{
        self.termination_detection_req = array(0, dtype=int)
        self.termination_detection_prequest = comm.Recv_init([self.termination_detection_req, MPI.INT], source=rank-1, tag=MSG_TERMINATION_DETECTION)
        self.computation_reqlist += [self.termination_detection_prequest]
        self.computation_reqhandlers += [self.termination_detection_received]
        #}}}

        self.reqlist = self.control_reqlist + self.computation_reqlist
        self.reqhandlers = self.control_reqhandlers + self.computation_reqhandlers

        #MPI buffer send space
        MPI.Attach_buffer(MPI.Alloc_mem(1024*1024)) #1Mb

        MPI.Prequest.Startall(self.control_reqlist + self.computation_reqlist)
        #print "Slave", rank, "awaiting your command"

        self.states_seen = 0
        self.running = True
        self.term_detect_color = WHITE
        while self.running:
            #if empty, wait for comm
            if self.waiting == []:
                #If we have the term detect token, pass it on
                if self.has_term_detect_token:
                    self.termination_detection_received()
                i = MPI.Prequest.Waitany(self.reqlist)
                self.reqhandlers[i]()
                self.reqlist[i].Start()
            else:
                #Do control comm, if any
                (i, comm_todo) = MPI.Prequest.Testany(self.control_reqlist)
                if comm_todo:
                    self.control_reqhandlers[i]()
                    self.control_reqlist[i].Start()
                #Do computation comm, if any
                (i, comm_todo) = MPI.Prequest.Testany(self.computation_reqlist)
                if comm_todo:
                    self.computation_reqhandlers[i]()
                    self.computation_reqlist[i].Start()

                #Do a computation
                state = self.waiting.pop()
                #Is this the goal state?
                if self.succgen.goalcheck(state):
                    comm.Bsend([None, MPI.INT], dest=0, tag=MSG_GOAL_REACHED)

                for succ in self.succgen.successors(state):
                    #send state to home node
                    self.sendstate(succ)
                    self.states_seen += 1
    #}}}

    #Slave callbacks {{{
    def abort(self):
        print "Slave %d: Aborting!" % (rank)
        self.running = False

    def status(self):
        #print "Slave %d: reporting!" % (rank)
        msg = (len(self.waiting), len(self.passed_waiting), self.states_seen)
        comm.send(msg, dest=0, tag=MSG_STATUS)

    def state_received(self):
        state = pickle.loads(self.state_req.tostring())
        if not state in self.passed_waiting:
            self.waiting.append(state)
            self.passed_waiting.add(state)

    def termination_detection_received(self):
        token = self.termination_detection_req
        #print "received:", token, "self is:", self.term_detect_color, token == WHITE
        #If inactive, pass it on
        if len(self.waiting) == 0:
            self.has_term_detect_token = False
            if token == WHITE and self.term_detect_color == WHITE:
                tosend = array(WHITE, dtype=int)
                #print "Sending term_detect from", rank
                comm.Bsend([tosend, MPI.INT], dest=(rank+1)%size, tag=MSG_TERMINATION_DETECTION)
            else:
                tosend = array(BLACK, dtype=int)
                #print "Sending term_detect from", rank
                comm.Bsend([tosend, MPI.INT], dest=(rank+1)%size, tag=MSG_TERMINATION_DETECTION)
                self.term_detect_color = WHITE
        else:
            self.has_term_detect_token = True
    #}}}

if __name__ == "__main__":
    model = None
    if rank == 0:
        #we're master, load model, do stuff, distribute it and start
        model_file = open(sys.argv[1], "r")
        model = pyuppaal.from_xml(model_file)
        comm.bcast((model, ""), root=0)
        #print "master: ", model

        query = sys.argv[2] or ""

        #start model checking
        node = Node(model, query)
        node.master()
    else:
        #wait for model
        (model, query) = comm.bcast(None, root=0)
        #print "slave:", model

        #start model checking
        node = Node(model, query)
        node.slave()

# vim:ts=4:sw=4:expandtab
