
from copy import deepcopy
import copy
from ulp import lexer, parser, expression_parser
import operator

class Error(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

class IllegalStateException(Error):
    pass
class IllegalQueryException(Error):
    pass

class State:
    def __init__(self):
        self.vars = {}
        self.locs = {}
        self.clocks = {}
        self.hash = None

    def __getitem__(self, key):
        if key == 'locs':
            return self.locs
        try:
            return self.clocks[key]
        except KeyError:
            return self.vars[key]
    def __setitem__(self, key, value):
        self.hash = None
        if key in self.clocks.keys():
            if value == 0:
                #print 'Reseting clock %s' % (key)
                self.clocks[key] = value
            else:
                raise IllegalStateException("Assignment to clock %s not 0" % (key))
        else:
            self.vars[key] = value

    def __eq__(self, other):
        #print "Comparing", self.locs, other.locs
        return self.clocks == other.clocks and self.vars == other.vars and \
            self.locs == other.locs

    def __hash__(self):
        #Hash like a tuple, see http://effbot.org/zone/python-hash.htm for details
        self.hash = self.hash or \
            hash(tuple(self.locs.values() + self.vars.values() + self.clocks.values()))
        return self.hash
        

    def __str__(self):
        return "(" + str(self.locs) + ", " + str(self.vars) + ", " + str(self.clocks) + ")"

    def copy(self):
        """Returns a copy of this state"""
        a = State()
        a.locs = self.locs.copy()
        a.vars = self.vars.copy()
        a.clocks = self.clocks.copy()
        return a

class SuccessorGenerator:
    def __init__(self, model, query):
        self.model = model
        self.transitions = {}
        self.sync_transitions = {}
        self.templatename_to_idx = {}
        self.num_templates = len(self.model.templates)

        #calculate variables, clocks and channels
        self.variables = []
        self.clocks = []
        self.channels = []
        lex = lexer.lexer
        pars = parser.Parser(model.declaration, lex)

        last_type = None
        def visit_identifiers(node):
            global last_type
            if node.type == 'VarDecl':
                last_type = node.leaf.type
            elif node.type == 'Identifier':
                if last_type == 'TypeInt':
                    self.variables += [(node.leaf, 'int')]
                elif last_type == 'TypeClock':
                    #TODO calculate max constant
                    self.clocks += [(node.leaf, 10)]
                elif last_type == 'TypeChannel':
                    self.channels += [node.leaf]
        pars.AST.visit(visit_identifiers)

        #TODO local variables and clocks
        print self.variables, self.clocks

        #calculate transitions
        for (t, t_idx) in zip(self.model.templates, 
                            range(len(self.model.templates))):
            self.transitions[t_idx] = {}
            self.sync_transitions[t_idx] = {}
            self.templatename_to_idx[t.name] = t_idx

            for (loc, l_idx) in zip(t.locations,
                                range(len(t.locations))):
                self.transitions[t_idx][l_idx] = []
                self.sync_transitions[t_idx][l_idx] = {}
                for chan_name in self.channels:
                    self.sync_transitions[t_idx][l_idx][chan_name + "!"] = []
                    self.sync_transitions[t_idx][l_idx][chan_name + "?"] = []
                for tr in [tr for tr in t.transitions if tr.source == loc]:
                    target_idx = t.locations.index(tr.target)
                    guard_code = None
                    if tr.guard.value != '':
                        guard_code = compile(tr.guard.value, "<guard>", 'eval')

                    update_code = None
                    if tr.assignment.value != '':
                        update_code = compile(tr.assignment.value, "<update>", 'exec')

                    list_curtrans = [(target_idx, guard_code, update_code, 
                        tr.synchronisation.value)]

                    self.transitions[t_idx][l_idx] += list_curtrans

                    if tr.synchronisation.value != "":
                        self.sync_transitions[t_idx][l_idx][tr.synchronisation.value] += list_curtrans

        #print "Transitions:"
        #print self.transitions





        # SHORTCIRCUIT ===============================
        if query == "":
            self.goal_locs = {}
            self.goal_temp_idxs = []
            return
        #=============================================
            

        #Parse query, and compile goal status query
        query_type = query.strip()[:3]
        query = (query.strip()[3:]).strip()

        if query_type != "E<>":
            raise IllegalQueryException(
                'Sorry, only reachability ("E<>") supported for now')
        query_ast = expression_parser.parser.parse(query)

        def expression_to_python(node):
            if node.type == 'and':
                return expression_to_python(node.children[0]) + \
                    " and " + expression_to_python(node.children[1])
            elif node.type == 'Identifier':
                i_name = node.leaf
                if i_name in [name for (name, _) in self.variables] or \
                    i_name in [name for (name, _) in self.clocks]:
                    return node.leaf
                raise IllegalQueryException('Unknown identifier ' + i_name)
            elif node.type == 'FieldAccess':
                #XXX when do we not have an identifier here?
                assert node.children[0].type == 'Identifier'
                #Assume "Template."
                t_idx = self.templatename_to_idx[node.children[0].leaf]
                template = self.model.templates[t_idx]
                f_name = node.children[1].leaf

                #case: Template.var | Template.clock
                if f_name in [name for (name, _) in self.variables]:
                    return f_name
                if f_name in [name for (name, _) in self.clocks]:
                    return f_name

                #case: Template.location
                for (loc, loc_idx) in zip(template.locations,
                    range(len(template.locations))):
                    if f_name == loc.name.value:
                        return "locs[%d] == %d" % (t_idx, loc_idx)

                raise IllegalQueryException('Unknown identifier ' + 
                    node.children[0].leaf + "." + node.children[1].leaf)


            return ""

        self.goal_func_python = expression_to_python(query_ast)
        print 'Goal func:', self.goal_func_python
        self.goal_expr = compile(self.goal_func_python, "<goal func>", 'eval')

    def successors(self, state):
        #Take an active transition
        for t_idx in xrange(self.num_templates):
            for (target_idx, guard, update, sync) in self.transitions[t_idx][state.locs[t_idx]]:
                #Evaluate guard
                #print "Evaluating guard: ", guard, "on", state.vars, state.clocks
                if guard == None or eval(guard, {}, state):
                    #If synchronisation, find trans to sync with
                    if sync != '':
                        if sync[-1] == '?':
                            #Only look for matching pair to ! syncs
                            continue
                        brother_sync = sync[:-1] + '?'
                        for t2_idx in xrange(self.num_templates):
                            for (target2_idx, guard2, update2, sync2) in self.sync_transitions[t2_idx][state.locs[t2_idx]][brother_sync]:
                                if guard2 == None or eval(guard2, {}, state):
                                    #sync found
                                    a = state.copy()
                                    a.locs[t_idx] = target_idx
                                    a.locs[t2_idx] = target2_idx
                                    #XXX, handle if update and update2 are overlapping
                                    if update != None:
                                        exec update in {}, a
                                    if update2 != None:
                                        exec update2 in {}, a
                                    yield a
                    #Not synchronising
                    else:
                        a = state.copy()
                        a.locs[t_idx] = target_idx
                        #Execute update
                        #print "Executing update", update, "on", a.vars
                        if update != None:
                            exec update in {}, a
                        #print "Result:", a.vars
                        yield a
        #Delay 1 time unit
        a = state.copy()
        delay_has_effect = False
        for (c, cmax) in self.clocks:
            if a.clocks[c] <= cmax:
                a.clocks[c] = a.clocks[c]+1
                delay_has_effect = True
        if delay_has_effect:
            yield a

    def get_initialstate(self):
        #calculate initial state
        initstate = State()
        for (t, idx) in zip(self.model.templates, 
                            range(len(self.model.templates))):
            #Get initial location
            initstate.locs[idx] = t.locations.index(t.initlocation)
            #Set initial vars
            for (var, vartype) in self.variables:
                initstate[var] = 0
            #Set initial clocks
            for (c, cmax) in self.clocks:
                initstate.clocks[c] = 0

        print "Initial:", initstate.locs, initstate.vars
        return initstate

    def goalcheck(self, state):
        if eval(self.goal_expr, {}, state):
            #print "Goal reached:", state.locs, self.goal_func_python
            return True
        return False

# vim:ts=4:sw=4:expandtab
